"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DefaultHandlerRegistry = void 0;
const error_1 = require("./error");
const handler_1 = require("./handler");
class DefaultHandlerRegistry {
    constructor() {
        this.registry = {};
        this.unhandledMessages = [];
        this.handlerResolvers = [];
    }
    registerCustom(handler, customResolver) {
        this.handlerResolvers.push({
            handler,
            messageType: undefined,
            resolver: customResolver.resolveWith,
            topicIdentifier: customResolver.topicIdentifier
        });
    }
    register(messageType, handler) {
        const message = new messageType();
        if (!('$name' in message)) {
            throw new error_1.SystemMessageMissingResolver(messageType);
        }
        const messageName = message.$name;
        if (!this.registry[messageName]) {
            // Register that the message will have subscriptions
            this.registry[messageName] = {
                messageType,
                handlers: []
            };
        }
        const handlerNameAlreadyRegistered = this.registry[messageName].handlers.some(registeredHandler => registeredHandler === handler);
        if (handlerNameAlreadyRegistered) {
            throw new error_1.HandlerAlreadyRegistered(handler.name);
        }
        this.registry[messageName].handlers.push(handler);
    }
    get(loggerFactory, message) {
        const logger = loggerFactory('@node-ts/bus-core:handler-registry');
        logger.debug('Getting handlers for message', { msg: message });
        const customHandlers = this.resolveCustomHandlers(message);
        const messageName = '$name' in message ? message.$name : undefined;
        const noHandlersForMessage = !customHandlers.length &&
            (!messageName || !(messageName in this.registry));
        if (noHandlersForMessage) {
            const warnNoHandlersForMessageType = messageName && !this.unhandledMessages.some(m => m === messageName);
            if (warnNoHandlersForMessageType) {
                this.unhandledMessages.push(messageName);
                logger.error(`No handlers were registered for message`, {
                    messageName,
                    help: `This could mean that either the handlers haven't been registered with bootstrap.registerHandler(),` +
                        ` or that the underlying transport is subscribed to messages that aren't handled and should be removed.`
                });
            }
            logger.debug('No handlers found for message', { msg: message });
            return [];
        }
        const messageHandlers = messageName && this.registry[messageName]
            ? this.registry[messageName].handlers
            : [];
        logger.debug('Found handlers for message', {
            msg: message,
            numMessageHandlers: messageHandlers.length,
            numCustomHandlers: customHandlers.length
        });
        const result = [...messageHandlers, ...customHandlers];
        return result;
    }
    getMessageNames() {
        return Object.keys(this.registry);
    }
    getMessageConstructor(messageName) {
        if (!(messageName in this.registry)) {
            return undefined;
        }
        return this.registry[messageName].messageType;
    }
    getExternallyManagedTopicIdentifiers() {
        return this.handlerResolvers
            .map(resolver => resolver.topicIdentifier)
            .filter(topicArn => !!topicArn);
    }
    getResolvers() {
        return this.handlerResolvers;
    }
    reset() {
        this.registry = {};
    }
    getClassHandlers() {
        const customHandlers = this.handlerResolvers
            .map(handlerResolver => handlerResolver.handler)
            .filter(handler_1.isClassHandler);
        const regularHandlers = Object.values(this.registry)
            .map(h => h.handlers)
            .flat()
            .filter(handler_1.isClassHandler);
        return [...customHandlers, ...regularHandlers];
    }
    /**
     * Returns a list of handlers that have been matched via custom resolvers for a message
     * @param message A message that has been fetched from the bus. Note that this may or may not
     * adhere to @node-ts/bus-messages/Message contracts, and could be any externally generated message.
     */
    resolveCustomHandlers(message) {
        return this.handlerResolvers
            .filter(handlerResolver => handlerResolver.resolver(message))
            .map(handlerResolver => handlerResolver.handler);
    }
}
exports.DefaultHandlerRegistry = DefaultHandlerRegistry;
