"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BusConfiguration = void 0;
const error_1 = require("../error");
const handler_1 = require("../handler");
const handler_2 = require("../handler/handler");
const logger_1 = require("../logger");
const retry_strategy_1 = require("../retry-strategy");
const serialization_1 = require("../serialization");
const message_serializer_1 = require("../serialization/message-serializer");
const transport_1 = require("../transport");
const util_1 = require("../util");
const persistence_1 = require("../workflow/persistence");
const workflow_registry_1 = require("../workflow/registry/workflow-registry");
const bus_instance_1 = require("./bus-instance");
const error_2 = require("./error");
class BusConfiguration {
    constructor() {
        this.concurrency = 1;
        this.workflowRegistry = new workflow_registry_1.WorkflowRegistry();
        this.handlerRegistry = new handler_1.DefaultHandlerRegistry();
        this.loggerFactory = logger_1.defaultLoggerFactory;
        this.serializer = new serialization_1.JsonSerializer();
        this.persistence = new persistence_1.InMemoryPersistence();
        this.messageReadMiddlewares = new util_1.MiddlewareDispatcher();
        this.retryStrategy = new retry_strategy_1.DefaultRetryStrategy();
        this.sendOnly = false;
        this.interruptSignals = ['SIGINT', 'SIGTERM'];
    }
    /**
     * Constructs an instance of a bus from the configuration
     */
    build() {
        var _a;
        if (!!this.busInstance) {
            throw new error_2.BusAlreadyInitialized();
        }
        const classHandlers = this.handlerRegistry.getClassHandlers();
        if (!this.container && classHandlers.length) {
            throw new error_1.ContainerNotRegistered(classHandlers[0].constructor.name);
        }
        const coreDependencies = {
            container: this.container,
            handlerRegistry: this.handlerRegistry,
            loggerFactory: this.loggerFactory,
            serializer: this.serializer,
            messageSerializer: new message_serializer_1.MessageSerializer(this.serializer, this.handlerRegistry),
            retryStrategy: this.retryStrategy,
            interruptSignals: this.interruptSignals
        };
        if (!this.sendOnly) {
            (_a = this.persistence) === null || _a === void 0 ? void 0 : _a.prepare(coreDependencies);
            this.workflowRegistry.prepare(coreDependencies, this.persistence);
        }
        const transport = this.configuredTransport || new transport_1.InMemoryQueue();
        transport.prepare(coreDependencies);
        this.busInstance = new bus_instance_1.BusInstance(transport, this.concurrency, this.workflowRegistry, coreDependencies, this.messageReadMiddlewares, this.handlerRegistry, this.container, this.sendOnly, this.receiver);
        return this.busInstance;
    }
    /**
     * Configure the bus to only send messages and not receive them. No queues or subscriptions will be created for
     * this service.
     */
    asSendOnly() {
        this.sendOnly = true;
        return this;
    }
    withHandler(...handler) {
        if (!!this.busInstance) {
            throw new error_2.BusAlreadyInitialized();
        }
        for (const handlerToAdd of handler) {
            if ('messageHandler' in handlerToAdd) {
                this.handlerRegistry.register(handlerToAdd.messageType, handlerToAdd.messageHandler);
            }
            else if ('messageResolver' in handlerToAdd) {
            }
            else if ((0, handler_2.isClassHandler)(handlerToAdd)) {
                const handlerInstance = new handlerToAdd();
                this.handlerRegistry.register(handlerInstance.messageType, handlerToAdd);
            }
        }
        return this;
    }
    /**
     * Registers a custom handler that receives messages from external systems, or messages that don't implement the
     * Message interface from @node-ts/bus-messages
     * @param messageHandler A handler that receives the custom message
     * @param customResolver A discriminator that determines if an incoming message should be mapped to this handler.
     */
    withCustomHandler(messageHandler, customResolver) {
        if (!!this.busInstance) {
            throw new error_2.BusAlreadyInitialized();
        }
        this.handlerRegistry.registerCustom(messageHandler, customResolver);
        return this;
    }
    /**
     * Register a workflow definition so that all of the messages it depends on will be subscribed to
     * and forwarded to the handlers inside the workflow
     */
    withWorkflow(...workflow) {
        if (!!this.busInstance) {
            throw new error_2.BusAlreadyInitialized();
        }
        workflow.forEach(workflowToRegister => this.workflowRegistry.register(workflowToRegister));
        return this;
    }
    /**
     * Configures Bus to use a different transport than the default MemoryQueue
     */
    withTransport(transportConfiguration) {
        if (!!this.busInstance) {
            throw new error_2.BusAlreadyInitialized();
        }
        this.configuredTransport = transportConfiguration;
        return this;
    }
    /**
     * Configures Bus to use a different logging provider than the default console logger
     */
    withLogger(loggerFactory) {
        if (!!this.busInstance) {
            throw new error_2.BusAlreadyInitialized();
        }
        this.loggerFactory = loggerFactory;
        return this;
    }
    /**
     * Configures Bus to use a different serialization provider. The provider is responsible for
     * transforming messages to/from a serialized representation, as well as ensuring all object
     * properties are a strong type
     */
    withSerializer(serializer) {
        if (!!this.busInstance) {
            throw new error_2.BusAlreadyInitialized();
        }
        this.serializer = serializer;
        return this;
    }
    /**
     * Configures Bus to use a different persistence provider than the default InMemoryPersistence provider.
     * This is used to persist workflow data and is unused if not using workflows.
     */
    withPersistence(persistence) {
        if (!!this.busInstance) {
            throw new error_2.BusAlreadyInitialized();
        }
        this.persistence = persistence;
        return this;
    }
    /**
     * Sets the message handling concurrency beyond the default value of 1, which will increase the number of messages
     * handled in parallel.
     */
    withConcurrency(concurrency) {
        if (concurrency < 1) {
            throw new Error('Invalid concurrency setting. Must be set to 1 or greater');
        }
        this.concurrency = concurrency;
        return this;
    }
    /**
     * Use a local dependency injection/IoC container to resolve handlers
     * and workflows.
     * @param container An adapter to an existing DI container to fetch class instances from
     */
    withContainer(container) {
        this.container = container;
        return this;
    }
    /**
     * Register optional middlewares that will run for each message that is polled from the transport
     * Note these middlewares only run when polling successfully pulls a message off the Transports queue
     * After all the user defined middlewares have registered.
     */
    withMessageReadMiddleware(messageReadMiddleware) {
        this.messageReadMiddlewares.use(messageReadMiddleware);
        return this;
    }
    /**
     * Configure @node-ts/bus to use a different retry strategy that determines delays between
     * retrying failed messages.
     * @default DefaultRetryStrategy
     */
    withRetryStrategy(retryStrategy) {
        this.retryStrategy = retryStrategy;
        return this;
    }
    /**
     * Register additional signals that will cause the bus to gracefully shutdown
     * @default [SIGINT, SIGTERM]
     */
    withAdditionalInterruptSignal(...signals) {
        this.interruptSignals = Array.from(new Set([...this.interruptSignals, ...signals]));
        return this;
    }
    /**
     * Register a receiving mechanism that will be used to receive messages and deliver
     * them to the dispatcher.
     *
     * Usually the bus will connect to a transport and receive messages directly. However
     * a different receiver plugin can be used that will become responsible for this instead.
     *
     * Once the bus is configured, messages can be received by passing the received message into bus.receive().
     *
     * @param receiver The receiver mechanism to use, or `undefined` to use the default behaviour
     */
    withReceiver(receiver) {
        this.receiver = receiver;
        return this;
    }
}
exports.BusConfiguration = BusConfiguration;
