"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BusInstance = void 0;
const tslib_1 = require("tslib");
const util_1 = require("../util");
const handler_1 = require("../handler");
const serialize_error_1 = require("serialize-error");
const bus_state_1 = require("./bus-state");
const message_handling_context_1 = require("../message-handling-context");
const error_1 = require("../error");
const uuid_1 = require("uuid");
const error_2 = require("./error");
const als_1 = tslib_1.__importDefault(require("alscontext/dist/als/als"));
const message_lifecycle_context_1 = require("../message-lifecycle-context");
const throat_1 = tslib_1.__importDefault(require("throat"));
const EMPTY_QUEUE_SLEEP_MS = 500;
class BusInstance {
    constructor(transport, concurrency, workflowRegistry, coreDependencies, messageReadMiddleware, handlerRegistry, container, sendOnly, receiver) {
        this.transport = transport;
        this.concurrency = concurrency;
        this.workflowRegistry = workflowRegistry;
        this.coreDependencies = coreDependencies;
        this.messageReadMiddleware = messageReadMiddleware;
        this.handlerRegistry = handlerRegistry;
        this.container = container;
        this.sendOnly = sendOnly;
        this.receiver = receiver;
        /**
         * Emitted before a command is sent to the transport
         */
        this.beforeSend = new util_1.TypedEmitter();
        /**
         * Emitted before an event is published to the transport
         */
        this.beforePublish = new util_1.TypedEmitter();
        /**
         * Emitted after a command has been sent to the transport
         */
        this.afterSend = new util_1.TypedEmitter();
        /**
         * Emitted after an event has been published to the transport
         */
        this.afterPublish = new util_1.TypedEmitter();
        /**
         * Emitted when an error occurs during message handling
         */
        this.onError = new util_1.TypedEmitter();
        /**
         * Emitted immediately after a message has been received from the transport
         */
        this.afterReceive = new util_1.TypedEmitter();
        /**
         * Emitted before a message is dispatched to handlers
         */
        this.beforeDispatch = new util_1.TypedEmitter();
        /**
         * Emitted after a message has been dispatched and completed all handler invocations
         */
        this.afterDispatch = new util_1.TypedEmitter();
        this.internalState = bus_state_1.BusState.Stopped;
        this.runningWorkerCount = 0;
        this.isInitialized = false;
        this.outbox = new als_1.default();
        /**
         * The final middleware that runs, after all the useBeforeHandleNextMessage middlewares have completed
         * It dispatches a message that has been polled from the queue
         * and deletes the message from the transport
         */
        this.handleNextMessagePolled = (message, next) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.dispatchMessageToHandlers(message.domainMessage, message.attributes);
            const { messageReturnedToQueue } = message_lifecycle_context_1.messageLifecycleContext.get();
            if (messageReturnedToQueue) {
                this.logger.debug('Message was returned to queue by a handler and will not be deleted', { message });
                // Receivers assume that the the host is responsible for deleting successful messages
            }
            else if (!this.receiver) {
                yield this.transport.deleteMessage(message);
            }
            return next();
        });
        this.logger = coreDependencies.loggerFactory('@node-ts/bus-core:service-bus');
        this.messageReadMiddleware.useFinal(this.handleNextMessagePolled);
    }
    /**
     * Receive one or more messages to dispatch directly to handlers. This can only be called when a Receiver
     * has been configured using Bus.configure().withReceiver()
     */
    receive(message) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this.receiver) {
                throw new error_2.InvalidOperation('Cannot use handler when a Receiver is not set. Use Bus.configure().withReceiver() to set a Receiver.');
            }
            this.logger.info('Received messages to process');
            const messagesReceived = yield this.receiver.receive(message, this.coreDependencies.messageSerializer);
            const messagesToDispatch = Array.isArray(messagesReceived)
                ? messagesReceived
                : [messagesReceived];
            this.logger.debug('Parsed messages from receiver', {
                numMessages: messagesToDispatch.length
            });
            // Throttle back to concurrency, since batch sizes can be far beyond this limit.
            const throttle = (0, throat_1.default)(this.concurrency);
            yield Promise.all(messagesToDispatch.map(message => throttle(() => this.handleReceivedMessage(message))));
            this.logger.debug('All received messages dispatched to handlers');
        });
    }
    /**
     * Initializes the bus with the provided configuration. This must be called before `.start()`
     *
     * @throws InvalidOperation if the bus has already been initialized
     */
    initialize() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.logger.debug('Initializing bus');
            if (this.isInitialized) {
                throw new error_2.InvalidOperation('Bus has already been initialized');
            }
            if (!this.sendOnly) {
                yield this.workflowRegistry.initialize(this.handlerRegistry, this.container);
            }
            if (this.transport.connect) {
                yield this.transport.connect({
                    concurrency: this.concurrency
                });
            }
            if (this.transport.initialize) {
                yield this.transport.initialize({
                    handlerRegistry: this.handlerRegistry,
                    sendOnly: this.sendOnly
                });
            }
            this.subscribeToInterruptSignals(this.coreDependencies.interruptSignals);
            this.isInitialized = true;
            this.logger.debug('Bus initialized', {
                sendOnly: this.sendOnly,
                registeredMessages: this.handlerRegistry.getMessageNames()
            });
        });
    }
    /**
     * Publishes an event to the transport
     * @param event An event to publish
     * @param messageAttributes A set of attributes to attach to the outgoing message when published
     */
    publish(event_1) {
        return tslib_1.__awaiter(this, arguments, void 0, function* (event, messageAttributes = {}) {
            this.logger.debug('Publishing event', { event, messageAttributes });
            const attributes = this.prepareTransportOptions(messageAttributes);
            this.beforePublish.emit({ event, attributes });
            if (message_handling_context_1.messageHandlingContext.isInHandlerContext) {
                // Add message to outbox and only send when the handler resolves
                const outbox = this.outbox.get('outbox');
                outbox.push({ event, attributes });
                this.outbox.set('outbox', outbox);
            }
            else {
                return this.transport.publish(event, attributes);
            }
        });
    }
    /**
     * Sends a command to the transport
     * @param command A command to send
     * @param messageAttributes A set of attributes to attach to the outgoing message when sent
     */
    send(command_1) {
        return tslib_1.__awaiter(this, arguments, void 0, function* (command, messageAttributes = {}) {
            this.logger.debug('Sending command', { command, messageAttributes });
            const attributes = this.prepareTransportOptions(messageAttributes);
            this.beforeSend.emit({ command, attributes });
            if (message_handling_context_1.messageHandlingContext.isInHandlerContext) {
                // Add message to outbox and only send when the handler resolves
                const outbox = this.outbox.get('outbox');
                outbox.push({ command, attributes });
                this.outbox.set('outbox', outbox);
            }
            else {
                return this.transport.send(command, attributes);
            }
        });
    }
    /**
     * Instructs the bus that the current message being handled cannot be processed even with
     * retries and instead should immediately be routed to the dead letter queue
     * @throws FailMessageOutsideHandlingContext if called outside a message handling context
     */
    failMessage() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const message = message_handling_context_1.messageHandlingContext.get();
            if (!message) {
                throw new error_1.FailMessageOutsideHandlingContext();
            }
            this.logger.debug('Failing message', { message });
            return this.transport.fail(message);
        });
    }
    /**
     * Instructs that the current message should be returned to the queue for retry.
     * @throws ReturnMessageOutsideHandlingContext if called outside a message handling context
     */
    returnMessage() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const context = message_lifecycle_context_1.messageLifecycleContext.get();
            const message = message_handling_context_1.messageHandlingContext.get();
            if (!context || !message) {
                throw new error_1.ReturnMessageOutsideHandlingContext();
            }
            message_lifecycle_context_1.messageLifecycleContext.set(Object.assign(Object.assign({}, context), { messageReturnedToQueue: true }));
            this.logger.debug('Returning message', { message });
            return this.transport.returnMessage(message);
        });
    }
    /**
     * Instructs the bus to start reading messages from the underlying service queue
     * and dispatching to message handlers.
     *
     * @throws InvalidOperation if the bus is configured to be send-only
     * @throws InvalidOperation if the bus has not been initialized
     * @throws InvalidOperation if the bus has a receiver set
     * @throws InvalidBusState if the bus is already started or in a starting state
     */
    start() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this.sendOnly) {
                throw new error_2.InvalidOperation('Cannot start a send-only bus');
            }
            if (!this.isInitialized) {
                throw new error_2.InvalidOperation('Bus must be initialized before starting');
            }
            if (this.receiver) {
                throw new error_2.InvalidOperation('Cannot start a bus with a Receiver. Pass incoming messages for dispatch by invoking bus.receive()');
            }
            const startedStates = [bus_state_1.BusState.Started, bus_state_1.BusState.Starting];
            if (startedStates.includes(this.state)) {
                throw new error_2.InvalidBusState('Bus must be stopped before it can be started', this.state, [bus_state_1.BusState.Stopped, bus_state_1.BusState.Stopping]);
            }
            this.internalState = bus_state_1.BusState.Starting;
            this.logger.info('Bus starting...');
            if (this.transport.start) {
                yield this.transport.start();
            }
            this.internalState = bus_state_1.BusState.Started;
            for (var i = 0; i < this.concurrency; i++) {
                setTimeout(() => tslib_1.__awaiter(this, void 0, void 0, function* () { return this.applicationLoop(); }), 0);
            }
            this.logger.info(`Bus started with concurrency ${this.concurrency}`);
        });
    }
    /**
     * Stops a bus that has been started by `.start()`. This will wait for all running workers to complete
     * their current message handling contexts before returning.
     *
     * @throws InvalidBusState if the bus is already stopped or stopping
     */
    stop() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const stoppedStates = [bus_state_1.BusState.Stopped, bus_state_1.BusState.Stopping];
            if (stoppedStates.includes(this.state)) {
                throw new error_2.InvalidBusState('Bus must be started before it can be stopped', this.state, [bus_state_1.BusState.Started, bus_state_1.BusState.Started]);
            }
            this.internalState = bus_state_1.BusState.Stopping;
            this.logger.info('Bus stopping...');
            if (this.transport.stop) {
                yield this.transport.stop();
            }
            while (this.runningWorkerCount > 0) {
                yield (0, util_1.sleep)(10);
            }
            this.internalState = bus_state_1.BusState.Stopped;
            this.logger.info('Bus stopped');
        });
    }
    /**
     * Stops and disposes all resources allocated to the bus, as well as removing
     * all handler registrations.
     *
     * The bus instance can not be used after this has been called.
     */
    dispose() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.logger.info('Disposing bus instance...');
            if (![bus_state_1.BusState.Stopped, bus_state_1.BusState.Stopped].includes(this.state)) {
                yield this.stop();
            }
            if (this.transport.disconnect) {
                yield this.transport.disconnect();
            }
            if (this.transport.dispose) {
                yield this.transport.dispose();
            }
            yield this.workflowRegistry.dispose();
            this.coreDependencies.handlerRegistry.reset();
            this.logger.info('Bus instance disposed');
        });
    }
    /**
     * Gets the current state of a message-handling bus
     */
    get state() {
        return this.internalState;
    }
    applicationLoop() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.runningWorkerCount++;
            // Run the loop in a cls-hooked namespace to provide the message handling context to all async operations
            while (this.internalState === bus_state_1.BusState.Started) {
                const messageHandled = yield this.handleNextMessage();
                // Avoids locking up CPU when there are no messages to be processed
                if (!messageHandled) {
                    yield (0, util_1.sleep)(EMPTY_QUEUE_SLEEP_MS);
                }
            }
            this.runningWorkerCount--;
        });
    }
    handleNextMessage() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            try {
                const message = yield this.transport.readNextMessage();
                if (message) {
                    return this.handleReceivedMessage(message);
                }
            }
            catch (error) {
                this.logger.error('Failed to handle and dispatch message from transport', { error: (0, serialize_error_1.serializeError)(error) });
            }
            return false;
        });
    }
    handleReceivedMessage(message) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            try {
                Object.freeze(message);
                this.logger.debug('Message read from transport', { message });
                this.afterReceive.emit({ message });
                return yield message_handling_context_1.messageHandlingContext.run(message, () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                    try {
                        yield message_lifecycle_context_1.messageLifecycleContext.run({ messageReturnedToQueue: false }, () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                            yield this.messageReadMiddleware.dispatch(message);
                            this.afterDispatch.emit({
                                message: message.domainMessage,
                                attributes: message.attributes
                            });
                        }));
                        return true;
                    }
                    catch (error) {
                        this.logger.error('Message was unsuccessfully handled. Returning to queue.', {
                            busMessage: message,
                            error: (0, serialize_error_1.serializeError)(error)
                        });
                        this.onError.emit({
                            message: message.domainMessage,
                            error: error,
                            attributes: message.attributes,
                            rawMessage: message
                        });
                        // Receivers expect the host to return the message to the queue for retry
                        if (this.receiver) {
                            throw error;
                        }
                        yield this.transport.returnMessage(message);
                        return false;
                    }
                }), true);
            }
            catch (error) {
                this.logger.error('Failed to handle and dispatch message from transport', { error: (0, serialize_error_1.serializeError)(error) });
                // Receivers expect the host to return the message to the queue for retry
                if (this.receiver) {
                    throw error;
                }
            }
            return false;
        });
    }
    dispatchMessageToHandlers(message, messageAttributes) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const handlers = this.coreDependencies.handlerRegistry.get(this.coreDependencies.loggerFactory, message);
            if (handlers.length === 0) {
                this.logger.error(`No handlers registered for message. Message will be discarded`, { messageName: message.$name });
                return;
            }
            const handlersToInvoke = handlers.map(handler => this.dispatchMessageToHandler(message, messageAttributes, handler));
            this.beforeDispatch.emit({
                message,
                attributes: messageAttributes,
                handlers
            });
            const handlerResults = yield Promise.allSettled(handlersToInvoke);
            const failedHandlers = handlerResults.filter(r => r.status === 'rejected');
            if (failedHandlers.length) {
                const reasons = failedHandlers.map(h => h.reason);
                throw new handler_1.HandlerDispatchRejected(reasons);
            }
            this.logger.debug('Message dispatched to all handlers', {
                message,
                numHandlers: handlersToInvoke.length
            });
        });
    }
    prepareTransportOptions(clientOptions) {
        const handlingContext = message_handling_context_1.messageHandlingContext.get();
        const messageAttributes = {
            // The optional operator? decided not to work here
            correlationId: clientOptions.correlationId ||
                (handlingContext
                    ? handlingContext.attributes.correlationId
                    : undefined) ||
                (0, uuid_1.v4)(),
            attributes: clientOptions.attributes || {},
            stickyAttributes: Object.assign(Object.assign({}, (handlingContext ? handlingContext.attributes.stickyAttributes : {})), clientOptions.stickyAttributes)
        };
        this.logger.debug('Prepared transport options', { messageAttributes });
        return messageAttributes;
    }
    dispatchMessageToHandler(message, attributes, handler) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            let handlerCallback;
            if ((0, handler_1.isClassHandler)(handler)) {
                const classHandler = handler;
                let handlerInstance;
                try {
                    const handlerInstanceFromContainer = this.coreDependencies.container.get(classHandler, {
                        message,
                        messageAttributes: attributes
                    });
                    if (handlerInstanceFromContainer instanceof Promise) {
                        handlerInstance = yield handlerInstanceFromContainer;
                    }
                    else {
                        handlerInstance = handlerInstanceFromContainer;
                    }
                    if (!handlerInstance) {
                        throw new Error('Container failed to resolve an instance.');
                    }
                }
                catch (e) {
                    throw new error_1.ClassHandlerNotResolved(e.message);
                }
                handlerCallback = () => tslib_1.__awaiter(this, void 0, void 0, function* () { return handlerInstance.handle(message, attributes); });
            }
            else {
                const fnHandler = handler;
                handlerCallback = () => tslib_1.__awaiter(this, void 0, void 0, function* () { return fnHandler(message, attributes); });
            }
            yield this.outbox.run({ outbox: [] }, () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                yield handlerCallback();
                // Flush outboxed messages for the handler
                const outboxedMessages = this.outbox.get('outbox');
                if (outboxedMessages && outboxedMessages.length > 0) {
                    // In case of a large number of messages to send, use a worker pool to dispatch so that we don't blow out heap usage
                    const dispatchWorkerCount = Math.min(outboxedMessages.length, 10);
                    const workers = new Array(dispatchWorkerCount)
                        .fill(undefined)
                        .map(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
                        while (true) {
                            const messageToSend = outboxedMessages.shift();
                            if (messageToSend === undefined) {
                                break;
                            }
                            const { command, event, attributes } = messageToSend;
                            if (command) {
                                yield this.transport.send(command, attributes);
                                this.afterSend.emit({ command, attributes });
                            }
                            else if (event) {
                                yield this.transport.publish(event, attributes);
                                this.afterPublish.emit({ event, attributes });
                            }
                        }
                    }));
                    yield Promise.all(workers);
                }
            }));
        });
    }
    /**
     * Subscribes to the interrupt signals to gracefully stop the bus
     */
    subscribeToInterruptSignals(signals) {
        if (this.sendOnly) {
            // Only applies to message handling buses
            return;
        }
        const startedStates = [bus_state_1.BusState.Started, bus_state_1.BusState.Starting];
        signals.forEach(signal => {
            process.on(signal, () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                if (!startedStates.includes(this.state)) {
                    // No need to stop a non-started bus
                    return;
                }
                this.logger.info(`Received ${signal} signal. Stopping bus...`);
                yield this.stop();
            }));
        });
    }
}
exports.BusInstance = BusInstance;
