"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.toTransportMessage = exports.InMemoryQueue = void 0;
const tslib_1 = require("tslib");
const stream_1 = require("stream");
const default_in_memory_queue_configuration_1 = require("./default-in-memory-queue-configuration");
/**
 * An in-memory message queue. This isn't intended for production use as all messages
 * are kept in memory and hence will be wiped when the application or host restarts.
 *
 * There are however legitimate uses for in-memory queues such as decoupling of non-mission
 * critical code inside of larger applications; so use at your own discretion.
 */
class InMemoryQueue {
    constructor(memoryQueueConfiguration = new default_in_memory_queue_configuration_1.DefaultInMemoryQueueConfiguration()) {
        this.memoryQueueConfiguration = memoryQueueConfiguration;
        this.queue = [];
        this.queuePushed = new stream_1.EventEmitter();
        this._deadLetterQueue = [];
    }
    prepare(coreDependencies) {
        this.coreDependencies = coreDependencies;
        this.logger = coreDependencies.loggerFactory('@node-ts/bus-core:in-memory-queue');
    }
    initialize() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.messagesWithHandlers = {};
            this.coreDependencies.handlerRegistry
                .getMessageNames()
                .forEach(messageName => (this.messagesWithHandlers[messageName] = {}));
        });
    }
    dispose() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this.queue.length > 0) {
                this.logger.warn('In-Memory queue being shut down, all messages will be lost.', { queueSize: this.queue.length });
            }
        });
    }
    publish(event, messageOptions) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.addToQueue(event, messageOptions);
        });
    }
    send(command, messageOptions) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.addToQueue(command, messageOptions);
        });
    }
    fail(transportMessage) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.sendToDeadLetterQueue(transportMessage);
        });
    }
    readNextMessage() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.logger.debug('Reading next message', {
                depth: this.depth,
                numberMessagesVisible: this.numberMessagesVisible
            });
            return new Promise(resolve => {
                const onMessageEmitted = () => {
                    unsubscribeEmitter();
                    clearTimeout(timeoutToken);
                    resolve(getNextMessage());
                };
                this.queuePushed.on('pushed', onMessageEmitted);
                const unsubscribeEmitter = () => this.queuePushed.off('pushed', onMessageEmitted);
                // Immediately returns the next available message, or undefined if none are available
                const getNextMessage = () => {
                    const availableMessages = this.queue.filter(m => !m.raw.inFlight);
                    if (availableMessages.length === 0) {
                        this.logger.debug('No messages available in queue');
                        return;
                    }
                    const message = availableMessages[0];
                    message.raw.inFlight = true;
                    return message;
                };
                const timeoutToken = setTimeout(() => {
                    unsubscribeEmitter();
                    resolve(undefined);
                }, this.memoryQueueConfiguration.receiveTimeoutMs);
                const nextMessage = getNextMessage();
                if (nextMessage) {
                    unsubscribeEmitter();
                    clearTimeout(timeoutToken);
                    resolve(nextMessage);
                }
                // Else wait for the timeout (empty return) or emitted event to return
            });
        });
    }
    deleteMessage(message) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const messageIndex = this.queue.indexOf(message);
            if (messageIndex < 0) {
                // actions like .failMessage() will cause the message to already be deleted
                this.logger.debug('Message already deleted', { message, messageIndex });
                return;
            }
            this.logger.debug('Deleting message', {
                queueDepth: this.depth,
                messageIndex
            });
            this.queue.splice(messageIndex, 1);
            this.logger.debug('Message Deleted', { queueDepth: this.depth });
        });
    }
    returnMessage(message) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const delay = this.coreDependencies.retryStrategy.calculateRetryDelay(message.raw.seenCount);
            message.raw.seenCount++;
            if (message.raw.seenCount >= this.memoryQueueConfiguration.maxRetries) {
                // Message retries exhausted, send to DLQ
                this.logger.info('Message retry limit exceeded, sending to dead letter queue', { message });
                yield this.sendToDeadLetterQueue(message);
            }
            else {
                setTimeout(() => {
                    message.raw.inFlight = false;
                }, delay);
            }
        });
    }
    /**
     * Gets the queue depth, which is the number of messages both queued and in flight
     */
    get depth() {
        return this.queue.length;
    }
    get deadLetterQueueDepth() {
        return this._deadLetterQueue.length;
    }
    /**
     * Returns all messages sitting in the dead letter queue. This is a copy of the queue
     * so mutative actions on this array will have no consequence.
     */
    get deadLetterQueue() {
        return [...this._deadLetterQueue];
    }
    /**
     * Gets the number of messages in the queue, excluding those in flight
     */
    get numberMessagesVisible() {
        return this.queue.filter(m => !m.raw.inFlight).length;
    }
    sendToDeadLetterQueue(message) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._deadLetterQueue.push(message);
            yield this.deleteMessage(message);
        });
    }
    addToQueue(message, messageOptions = { attributes: {}, stickyAttributes: {} }) {
        if (this.messagesWithHandlers[message.$name]) {
            const transportMessage = (0, exports.toTransportMessage)(message, messageOptions, false);
            this.queue.push(transportMessage);
            this.logger.debug('Added message to queue', {
                message,
                queueSize: this.queue.length
            });
        }
        else {
            this.logger.warn('Message was not sent as it has no registered handlers', { message });
        }
    }
}
exports.InMemoryQueue = InMemoryQueue;
const toTransportMessage = (message, messageOptions, isProcessing) => ({
    id: undefined,
    domainMessage: message,
    attributes: messageOptions,
    raw: {
        seenCount: 0,
        payload: message,
        inFlight: isProcessing
    }
});
exports.toTransportMessage = toTransportMessage;
