"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.InMemoryPersistence = void 0;
const tslib_1 = require("tslib");
const workflow_state_1 = require("../workflow-state");
const error_1 = require("./error");
/**
 * A non-durable in-memory persistence for storage and retrieval of workflow state. Before using this,
 * be warned that all workflow state will not survive a process restart or application shut down. As
 * such this should only be used for testing, prototyping or handling unimportant workflows.
 */
class InMemoryPersistence {
    constructor() {
        this.workflowState = {};
    }
    prepare(coreDependencies) {
        this.logger = coreDependencies.loggerFactory('@node-ts/bus-core:in-memory-persistence');
    }
    initializeWorkflow(workflowStateConstructor, _) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const name = new workflowStateConstructor().$name;
            this.workflowState[name] = [];
        });
    }
    getWorkflowState(workflowStateConstructor, messageMap, message, attributes, includeCompleted) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const filterValue = messageMap.lookup(message, attributes);
            if (!filterValue) {
                return [];
            }
            const workflowStateName = new workflowStateConstructor().$name;
            const workflowState = this.workflowState[workflowStateName];
            if (!workflowState) {
                throw new error_1.WorkflowStateNotInitialized('Workflow state not initialized');
            }
            return workflowState.filter(data => (includeCompleted || data.$status === workflow_state_1.WorkflowStatus.Running) &&
                data[messageMap.mapsTo] === filterValue);
        });
    }
    saveWorkflowState(workflowState) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const workflowStateName = workflowState.$name;
            const existingWorkflowState = this.workflowState[workflowStateName];
            const existingItem = existingWorkflowState.find(d => d.$workflowId === workflowState.$workflowId);
            if (existingItem) {
                try {
                    Object.assign(existingItem, workflowState);
                }
                catch (err) {
                    this.logger.error('Unable to update data', { err });
                    throw err;
                }
            }
            else {
                existingWorkflowState.push(workflowState);
            }
        });
    }
    length(workflowStateConstructor) {
        return this.workflowState[workflowStateConstructor.prototype.constructor.NAME].length;
    }
}
exports.InMemoryPersistence = InMemoryPersistence;
