"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WorkflowRegistry = void 0;
const tslib_1 = require("tslib");
const workflow_state_1 = require("../workflow-state");
const uuid = tslib_1.__importStar(require("uuid"));
const workflow_1 = require("../workflow");
const error_1 = require("../persistence/error");
const error_2 = require("../error");
const message_handling_context_1 = require("../../message-handling-context");
const handler_1 = require("../../handler");
/**
 * A default lookup that will match a workflow by its id with the workflowId
 * stored in the sticky attributes
 */
const workflowLookup = {
    lookup: (_, attributes) => attributes.stickyAttributes.workflowId,
    mapsTo: '$workflowId'
};
/**
 * The central workflow registry that holds all workflows managed by the application. This includes
 *   - the list of workflows
 *   - what messages start the workflow
 *   - what messages are handled by each workflow
 * This registry is also responsible for dispatching messages to workflows as they are received.
 */
class WorkflowRegistry {
    constructor() {
        this.workflowRegistry = [];
        this.isInitialized = false;
        this.isInitializing = false;
    }
    prepare(coreDependencies, persistence) {
        this.logger = coreDependencies.loggerFactory('@node-ts/bus-core:workflow-registry');
        this.persistence = persistence;
    }
    register(workflow) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this.isInitialized) {
                throw new Error(`Attempted to register workflow (${workflow.prototype.constructor.name}) after workflows have been initialized`);
            }
            const duplicateWorkflowName = this.workflowRegistry.some(r => r.prototype.constructor.name === workflow.prototype.constructor.name);
            if (duplicateWorkflowName) {
                throw new Error(`Attempted to register two workflows with the same name (${workflow.prototype.constructor.name})`);
            }
            this.workflowRegistry.push(workflow);
        });
    }
    /**
     * Initialize all services that are used to support workflows. This registers all messages subscribed to
     * in workflows as handlers with the bus, as well as initializing the persistence service so that workflow
     * states can be stored.
     *
     * This should be called once as the application is starting.
     */
    initialize(handlerRegistry, container) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this.workflowRegistry.length === 0) {
                this.logger.info('No workflows registered, skipping workflow initialization.');
                return;
            }
            if (this.isInitialized || this.isInitializing) {
                throw new error_2.WorkflowAlreadyInitialized();
            }
            this.logger.info('Initializing workflows...', {
                numWorkflows: this.workflowRegistry.length
            });
            this.isInitializing = true;
            if (this.persistence.initialize) {
                this.logger.info('Initializing persistence...');
                yield this.persistence.initialize();
            }
            for (const WorkflowCtor of this.workflowRegistry) {
                this.logger.debug('Initializing workflow', {
                    workflow: WorkflowCtor.prototype.constructor.name
                });
                let workflowInstance;
                if (container) {
                    const workflowInstanceFromContainer = container.get(WorkflowCtor);
                    if (workflowInstanceFromContainer instanceof Promise) {
                        workflowInstance = yield workflowInstanceFromContainer;
                    }
                    else {
                        workflowInstance = workflowInstanceFromContainer;
                    }
                }
                else {
                    workflowInstance = new WorkflowCtor();
                }
                const mapper = new workflow_1.WorkflowMapper(WorkflowCtor);
                workflowInstance.configureWorkflow(mapper);
                if (!mapper.workflowStateCtor) {
                    throw new Error('Workflow state not provided. Use .withState()');
                }
                this.registerFnStartedBy(mapper, handlerRegistry, container);
                this.registerFnHandles(mapper, handlerRegistry, WorkflowCtor, container);
                const messageWorkflowMappings = Array.from(mapper.onWhen, ([_, onWhenHandler]) => onWhenHandler.customLookup || workflowLookup);
                yield this.persistence.initializeWorkflow(mapper.workflowStateCtor, messageWorkflowMappings);
                this.logger.debug('Workflow initialized', {
                    workflowName: WorkflowCtor.prototype.name
                });
            }
            this.workflowRegistry = [];
            this.isInitialized = true;
            this.isInitializing = false;
            this.logger.info('Workflows initialized');
        });
    }
    dispose() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const isPrepared = this.persistence !== undefined;
            if (!isPrepared) {
                // If the registry has not been prepared, then there is no logger or persistence available
                return;
            }
            this.logger.debug('Disposing workflow registry');
            try {
                if (this.persistence.dispose) {
                    yield this.persistence.dispose();
                }
            }
            catch (error) {
                if (error instanceof error_1.PersistenceNotConfigured) {
                    return;
                }
                throw error;
            }
        });
    }
    registerFnStartedBy(mapper, handlerRegistry, container) {
        this.logger.debug('Registering started by handlers for workflow', {
            numHandlers: mapper.onStartedBy.size
        });
        mapper.onStartedBy.forEach((options, messageConstructor) => handlerRegistry.register(messageConstructor, (message, messageAttributes) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.logger.debug('Starting new workflow instance', {
                workflow: options.workflowCtor,
                msg: message
            });
            const workflowState = this.createWorkflowState(mapper.workflowStateCtor);
            const immutableWorkflowState = Object.freeze(Object.assign({}, workflowState));
            const workflowContext = this.buildWorkflowHandlingContext(immutableWorkflowState);
            // Extend the current message handling context, and augment with workflow-specific context data
            yield message_handling_context_1.messageHandlingContext.run(workflowContext, () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                yield this.dispatchMessageToWorkflow(message, messageAttributes, options.workflowCtor, immutableWorkflowState, mapper.workflowStateCtor, options.workflowHandler, container);
            }), true);
        })));
    }
    registerFnHandles(mapper, handlerRegistry, workflowCtor, container) {
        this.logger.debug('Registering handles for workflow', {
            workflow: workflowCtor,
            numHandlers: mapper.onWhen.size
        });
        mapper.onWhen.forEach((handler, messageConstructor) => {
            // TODO implement outbound tagging of workflowId to stickyAttributes
            const messageMapping = handler.customLookup || workflowLookup;
            handlerRegistry.register(messageConstructor, (message, attributes) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                this.logger.debug('Getting workflow state for message handler', {
                    msg: message,
                    workflow: workflowCtor
                });
                const workflowState = yield this.persistence.getWorkflowState(mapper.workflowStateCtor, messageMapping, message, attributes, false);
                if (!workflowState.length) {
                    this.logger.error('No existing workflow state found for message. Ignoring.', { busMessage: message, attributes });
                    return;
                }
                const workflowHandlers = workflowState.map((state) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                    // Extend the current message handling context, and augment with workflow-specific context data
                    const immutableWorkflowState = Object.freeze(Object.assign({}, state));
                    const workflowContext = this.buildWorkflowHandlingContext(immutableWorkflowState);
                    yield message_handling_context_1.messageHandlingContext.run(workflowContext, () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                        yield this.dispatchMessageToWorkflow(message, attributes, workflowCtor, immutableWorkflowState, mapper.workflowStateCtor, handler.workflowHandler, container);
                    }), true);
                }));
                const handlerResults = yield Promise.allSettled(workflowHandlers);
                const failedHandlers = handlerResults.filter(r => r.status === 'rejected');
                if (failedHandlers.length) {
                    const reasons = failedHandlers.map(h => h.reason);
                    throw new handler_1.HandlerDispatchRejected(reasons);
                }
            }));
        });
    }
    createWorkflowState(workflowStateType) {
        const data = new workflowStateType();
        data.$version = 0;
        data.$status = workflow_state_1.WorkflowStatus.Running;
        data.$workflowId = uuid.v4();
        this.logger.debug('Created new workflow state', {
            workflowId: data.$workflowId,
            workflowStateType
        });
        return data;
    }
    /**
     * Creates a new handling context for a single workflow. This is used so
     * that the `workflowId` is attached to outgoing messages in sticky
     * attributes. This allows message chains to be automatically mapped
     * back to the workflow if handled.
     */
    buildWorkflowHandlingContext(workflowState) {
        this.logger.debug('Starting new workflow handling context', {
            workflowState
        });
        const handlingContext = message_handling_context_1.messageHandlingContext.get();
        const workflowHandlingContext = structuredClone(handlingContext);
        workflowHandlingContext.attributes.stickyAttributes.workflowId =
            workflowState.$workflowId;
        return workflowHandlingContext;
    }
    dispatchMessageToWorkflow(message, attributes, workflowCtor, immutableWorkflowState, workflowStateConstructor, workflowHandler, container) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.logger.debug('Dispatching message to workflow', {
                msg: message,
                workflow: workflowCtor
            });
            let workflow;
            if (container) {
                const workflowFromContainer = container.get(workflowCtor, {
                    message,
                    messageAttributes: attributes
                });
                if (workflowFromContainer instanceof Promise) {
                    workflow = yield workflowFromContainer;
                }
                else {
                    workflow = workflowFromContainer;
                }
            }
            else {
                workflow = new workflowCtor();
            }
            const handler = workflow[workflowHandler];
            // Invoke the workflow handler
            const workflowStateOutput = yield handler.bind(workflow)(message, immutableWorkflowState, attributes);
            const workflowName = workflowCtor.prototype.name;
            if (workflowStateOutput &&
                workflowStateOutput.$status === workflow_state_1.WorkflowStatus.Discard) {
                this.logger.debug('Workflow step is discarding state changes. State changes will not be persisted', { workflowId: immutableWorkflowState.$workflowId, workflowName });
            }
            else if (workflowStateOutput || immutableWorkflowState.$version === 0) {
                // Persist the original workflow state if nothing's returned from the workflow startedBy function
                const workflowStateToChange = workflowStateOutput !== null && workflowStateOutput !== void 0 ? workflowStateOutput : immutableWorkflowState;
                this.logger.debug('Changes detected in workflow state and will be persisted.', {
                    workflowId: immutableWorkflowState.$workflowId,
                    workflowName,
                    changes: workflowStateToChange
                });
                const updatedWorkflowState = Object.assign(new workflowStateConstructor(), immutableWorkflowState, workflowStateToChange);
                try {
                    yield this.persist(updatedWorkflowState);
                }
                catch (error) {
                    this.logger.warn('Error persisting workflow state', {
                        err: error,
                        workflow: workflowName
                    });
                    throw error;
                }
            }
            else {
                this.logger.debug('No changes detected in workflow state.', {
                    workflowId: immutableWorkflowState.$workflowId
                });
            }
        });
    }
    persist(data) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            try {
                yield this.persistence.saveWorkflowState(data);
                this.logger.debug('Workflow state saved', { data });
            }
            catch (err) {
                this.logger.error('Error persisting workflow state', { err });
                throw err;
            }
        });
    }
}
exports.WorkflowRegistry = WorkflowRegistry;
